# Plots reaction time data from the AutoIt script.
# This script is old and needs some revision, but it should work.
# Might require data from at least two testing sessions before it'll work.
# Need to set appropriate file paths below.
#
# Questions?  Contact Alex Chernavsky at alexc@aya.yale.edu

require(ggplot2)
require(scales)
require(Cairo)

cat('\f') # Clear the console by sending a form-feed character to it
rm(list = ls(all.names = TRUE))  # Clean up the workspace by removing all the old objects

setwd("C:/Users/username/OneDrive/R/reaction time/Current")
raw.data <- read.csv("C:/Users/username/OneDrive/R/reaction time/Current/all_trials.csv")


raw.data$Posix.Time <- as.POSIXct(raw.data$Local_Time, format = "%m/%d/%Y %H:%M")

only.valid.trials <- raw.data[which(raw.data$Use_It. == 1), ] # Keep only those rows that are valid trials

# Calculate the mean latencies and group them by session number (not date/time, at this point):
summary1 <- aggregate(only.valid.trials$Latency, by = list(only.valid.trials$Session_Number), FUN = mean)

# Aggregate the session numbers by the starting date/time for each session (in other words, the date/time of
# the first trial for that session).
# The start date/time will then be associated with the mean reaction time for that session.
summary2 <- aggregate(only.valid.trials$Posix.Time, by = list(only.valid.trials$Session_Number), FUN = min)

# Change the variable names (column names) to something more meaningful:
colnames(summary1)[1] <- "Session.Number"
colnames(summary1)[2] <- "Mean.Latency"
colnames(summary2)[1] <- "Session.Number"
colnames(summary2)[2] <- "Start.Time"

# Combine the two "summary" dataframes into one:
summary3 <- merge(summary1, summary2, by = "Session.Number")

# Make sure that the summary dataframe is in the proper order:
summary3 <- summary3[order(summary3$Session.Number), ]

# CairoWin()
p1 <- ggplot(summary3, aes(x = Start.Time, y = Mean.Latency)) +
  ggtitle("Mean Reaction Time by Date") + 
  theme(plot.margin        = unit(c(t = 7, r = 15, b = 7, l = 7), "points"),
        panel.border       = element_rect(fill = NA, color = "black"),
        panel.background   = element_blank(),
        panel.grid.minor.x = element_blank(),
        panel.grid.minor.y = element_blank(),
        plot.title         = element_text(size = 18, face = "bold", vjust = 1.5, hjust = 0.5),
        axis.title.x       = element_blank(),
        axis.title.y       = element_blank(),
        axis.text.x        = element_text(face = "bold", color = "black", size = 8, angle = -45, hjust = 0),
        axis.text.y        = element_text(face = "bold", color = "black", size = 11)) +  
  geom_point(shape = 19, size = 1.5, fill = "black") +
  geom_smooth(method = "loess", span = 0.13, se = FALSE, size = 1.2) + 
  coord_cartesian(ylim = c(290, 380)) +
  scale_x_datetime(breaks = date_breaks('month'), labels = date_format('%b-%Y'), expand = c(0.008,0)) +
  scale_y_continuous(breaks = c(seq(290, 380, 10)), labels = comma) +
  # Make the last data point red and big, so that it stands out more:
  geom_point(data=tail(summary3, n = 1), color = "red", size = 3)
print(p1)
ggsave(filename = "latency.png", dpi=95, type="cairo-png")


# Now plot the number of errors for each testing session:

# Keep only those rows that indicate that a mistake was made, and don't include correction trials:
mistakes <- raw.data[which(raw.data$Use_It. == 0 & regexpr("c", raw.data$Target) != -1), ]

# The "length" function counts the number of occurrences:
summary1 <- aggregate(mistakes$Use_It., by = list(mistakes$Session_Number), FUN = length)

colnames(summary1)[1] <- "Session.Number"
colnames(summary1)[2] <- "Wrong.Answers"

summary1 <- summary1[order(summary1$Session.Number), ]
summary3 <- merge(summary1, summary2, by = "Session.Number")

p2 <- ggplot(summary3, aes(Start.Time, Wrong.Answers)) +
  ggtitle("Number of Wrong Answers by Session Date") + 
  theme(plot.margin        = unit(c(t = 7, r = 15, b = 7, l = 7), "points"),
        panel.border       = element_rect(fill = NA, color = "black"),
        panel.background   = element_blank(),
        panel.grid.minor.x = element_blank(),
        panel.grid.minor.y = element_blank(),
        plot.title         = element_text(size = 16, face = "bold", vjust = 1.5, hjust = 0.5),
        axis.title.x       = element_blank(),
        axis.text.x        = element_text(face = "bold", color = "black", size = 8, angle = -45, hjust = 0),
        axis.title.y       = element_text(face = "bold", color = "black", size = 11, vjust=0.25, margin = margin(t = 0, r = 10, b = 0, l = 0)), 
        axis.text.y        = element_text(face = "bold", color = "black", size = 11)) +  
  geom_point(shape = 19, size = 1.5, fill = "black") +
  geom_smooth(method = "loess", span = 0.13, se = FALSE, size = 1.2) + 
  coord_cartesian(ylim = c(0, 35)) +
  scale_x_datetime(breaks = date_breaks('month'), labels = date_format('%b-%Y'), expand = c(0.008,0)) +
  ylab("Number of Wrong Answers") +
  scale_y_continuous(breaks=c(seq(0, 35, 5)), labels=comma) +
  # Make the last data point red and big, so that it stands out more:
  geom_point(data=tail(summary3, n = 1), color = "red", size = 3)
print(p2)
ggsave(filename = "wrong_answers.png", dpi=95, type="cairo-png")
