#Region ;**** Directives created by AutoIt3Wrapper_GUI ****
#AutoIt3Wrapper_UseX64=y
#EndRegion ;**** Directives created by AutoIt3Wrapper_GUI ****
Opt("MustDeclareVars", 1)
#include <GUIConstantsEx.au3>
#include <Misc.au3>
#include <Date.au3>
#include <ButtonConstants.au3>
#include <EditConstants.au3>
#include <StaticConstants.au3>
#include <Array.au3>
; #include <WindowsConstants.au3>
#include "StringSize.au3"

; GUI Constants
Const $iGUIWidth = 450
Const $iGUIHeight = 207
Const $iSmallerFont = 20
Const $iMediumFont = 30
Const $iLargerFont = 40
Const $iDisplayDurationForNonRegularFeedbackGUI = 1000

; Program Constants
Const $iNumberOfTargets = 35  ; Fixed at 35 trials (5 presentations of each target digit, 2-8)

; Global Variables
Global $g_aSessionData[0][3] ; Columns: [0]=Target, [1]=Latency, [2]=Session_Number
Global $g_bJustCompletedCorrection = False
Global $g_iAvgReactionTimePastTwentySessions = 0

; Local Variables
Local $sOutputFileName  ; Name of output file that stores the data
Local $iMinTime  ; Minimum delay between trials (in milliseconds)
Local $iMaxTime  ; Maximum delay between trials (ms)
Local $iMinAllowedLatency  ; Minimum valid reaction time (ms)
Local $iMaxAllowedLatency  ; Maximum valid reaction time (ms)
Local $aiTargets[35]  ; Array contains the target digits (2-8, each appearing 5 times)
Local $avReactionTimeData[250][4]  ; User-generated data: Col 0=target, 1=latency, 2=correct/not, 3=timestamp
Local $sCondition  ; The user's "condition" for this test
Local $sSessionCode  ; Optional code for the session
Local $sOutputType  ; "WriteToFile" or "DisplayOnly"
Local $hGui  ; GUI handle
Local $hBegin  ; Start time when target is presented
; Local $hLabelMessage  ; GUI label for displaying messages
Local $iTotalTrialsConducted  ; Total trials done during a testing session (including incorrect ones)
Local $avResultsTemp[4]  ; Temporary vector for trial results
Local $aStringSize  ; For centering text in GUI
Local $iXPosition, $iYPosition  ; For positioning GUI elements
Local $bResultsAreValid  ; Validity check flag
Local $iNumberOfPastSessions  ; Number of past testing sessions
Local $bUsePercentiles = True
Local $iPercentileRank  ; Percentile rank for the user's latency in a given trial
Local $fFeedbackDisplayTime  ; Length of time (in milliseconds) to display feedback (latency, percentile) to the user
Local $iTrialState = "REGULAR"  ; States: "REGULAR", "CORRECTION"
Local $iCurrentTarget = 0
Local $iTotalValidTrialsCompleted = 0
Local $iDataIndex = 0

; Initialize the targets array - each digit 2-8 appears 5 times, assuming 35 total trials
Local $x = 0
For $digit = 2 To 8
    For $repeat = 1 To 5
        $aiTargets[$x] = $digit
        $x += 1
    Next
Next

; Shuffle the targets array
_ScrambleTargets($aiTargets)

HotKeySet("{ESC}", "_Terminate") ; Terminate the entire script if user hits ESC

; The following line introduces a 50 millisecond delay between keystrokes.  It doesn't (or shouldn't) affect
; the time before the first keystroke is sent.  If you don't have this, you get "bounce" in the GUI that detects
; keystrokes.  Anyway, it shouldn't affect the reaction time values.
AutoItSetOption("SendKeyDelay", 50)

_ReadParametersFromINIFile($sOutputFileName, $iMinTime, $iMaxTime, $iMinAllowedLatency, $iMaxAllowedLatency, $sOutputType, $fFeedbackDisplayTime)
_ConfirmParametersWithUser($sOutputFileName, $iMinTime, $iMaxTime, $iMinAllowedLatency, $iMaxAllowedLatency, $sOutputType, $sCondition, $sSessionCode, $fFeedbackDisplayTime)

$iNumberOfPastSessions = _GetNumberOfPastSessions($sOutputFileName)
If $iNumberOfPastSessions < 20 Then
	$bUsePercentiles = False
Else
	_GetDataFromPastSessions($sOutputFileName, $iNumberOfPastSessions)
EndIf

$hGui = GUICreate("Reaction Time Test", $iGUIWidth, $iGUIHeight, -1, -1, -1, -1) ; Display the GUI that will contain the targets and the feedback for the test-taker
GUISetFont($iSmallerFont)
GUISetState()

; Display initial instruction for the first trial
_WaitForUserToHitAFive($hGui, "." & @CRLF & @CRLF & """Esc"" aborts program at any point.", $iSmallerFont)

; Main loop runs until we have completed the required number of valid trials
While $iTotalValidTrialsCompleted < $iNumberOfTargets
    ; Determine the current target and display appropriate trial info
	If $iTrialState == "REGULAR" Then
		$iCurrentTarget = $aiTargets[$iTotalValidTrialsCompleted]
		If $iTotalValidTrialsCompleted > 0 OR $g_bJustCompletedCorrection Then
			_WaitForUserToHitAFive($hGui, @CRLF & "trial " & ($iTotalValidTrialsCompleted + 1) & " of " & $iNumberOfTargets & ".", $iMediumFont)
			$g_bJustCompletedCorrection = False  ; Reset the flag after using it
		EndIf
	Else ; CORRECTION state
		_WaitForUserToHitAFive($hGui, @CRLF & "correction trial" & ".", $iMediumFont)
	EndIf

    ; Present the target and get the user's response
    $avResultsTemp = _PresentTargetAndGetResponse($iCurrentTarget, $hGui)

    ; Record result regardless of success/failure
    $avReactionTimeData[$iDataIndex][0] = ($iTrialState == "CORRECTION") ? "c" : $avResultsTemp[0]
    $avReactionTimeData[$iDataIndex][1] = $avResultsTemp[1]
    $avReactionTimeData[$iDataIndex][2] = $avResultsTemp[2]
    $avReactionTimeData[$iDataIndex][3] = $avResultsTemp[3]

    ; Determine if trial was valid (correct response and within allowed time)
    Local $bValidResult = ($avResultsTemp[2] == 1) And ($avResultsTemp[1] <= $iMaxAllowedLatency) And ($avResultsTemp[1] >= $iMinAllowedLatency)

    ; Handle result based on current state
    If $iTrialState == "REGULAR" Then
        If $bValidResult Then
            ; Success in regular trial - show feedback and move to next target
            _ShowSuccessFeedback($avResultsTemp[1], $iCurrentTarget)
            $iTotalValidTrialsCompleted += 1
        Else
            ; Failure in regular trial - determine error type and show feedback
            _ShowFailureFeedback($avResultsTemp)
            ; Move to correction state
            $iTrialState = "CORRECTION"
        EndIf
    Else ; CORRECTION state
        If $avResultsTemp[2] == 1 Then ; Correct response in correction trial
            ; Success in correction - show feedback
            _ShowCorrectionSuccessFeedback()

			; Set flag to indicate we just completed a correction trial
			$g_bJustCompletedCorrection = True

            ; Return to regular trials
            $iTrialState = "REGULAR"

            ; Shuffle remaining targets to avoid immediate repetition
            If $iTotalValidTrialsCompleted < ($iNumberOfTargets - 1) Then
                _ScrambleTargets($aiTargets, $iTotalValidTrialsCompleted)
            EndIf
        Else
            ; Failure in correction - show feedback and stay in correction state
            _ShowCorrectionFailureFeedback()
            ; Stay in correction state (no state change)
        EndIf
    EndIf

    ; Increment data array index
    $iDataIndex += 1
WEnd

GUIDelete($hGui)
$iTotalTrialsConducted = $iDataIndex
_DisplaySummary($iNumberOfTargets, $iTotalTrialsConducted, $avReactionTimeData)
$bResultsAreValid = _CheckForValidResults($iNumberOfTargets, $iTotalTrialsConducted, $avReactionTimeData)
If ($bResultsAreValid = False) Then
    Exit
EndIf
_WriteReactionTimeData($sCondition, $iTotalTrialsConducted, $avReactionTimeData)
_UpdateINIFile($sOutputFileName, $iMinTime, $iMaxTime, $iMinAllowedLatency, $iMaxAllowedLatency, $sOutputType, $fFeedbackDisplayTime)
; End of main body


Func _ScrambleTargets(ByRef $aiArray, $iStart = 0)
    Local $temp
    Local $uBound = UBound($aiArray) - 1

    For $i = $iStart To $uBound
        Local $randomIndex = Random($iStart, $uBound, 1)
        ; Swap elements
        $temp = $aiArray[$i]
        $aiArray[$i] = $aiArray[$randomIndex]
        $aiArray[$randomIndex] = $temp
    Next
    Return 1
EndFunc


Func _GetDataFromPastSessions($sFilePath, $iMaxSessions)
    ; This function gets called only if there are at least 20 sessions of data stored in
    ; the output file. So we're not going to check for the lower bound. But if there are
    ; more than 60 sessions, we're going to consider only the last 60 of them for the purposes
	; of the percentile calculations.
	; This function also calculates the average reaction time for the past 20 sessions (we're
	; not going back beyond 20).
    If $iMaxSessions > 60 Then $iMaxSessions = 60

    ; Verify input file exists
    If Not FileExists($sFilePath) Then
        ConsoleWrite("Error: File '" & $sFilePath & "' not found!" & @CRLF)
        Return False
    EndIf

    SplashTextOn("Please Wait", "Please wait... initializing percentile data", 300, 80)
    ; Read the entire file at once
    Local $sFileContent = FileRead($sFilePath)
    If @error Or $sFileContent = "" Then
        ConsoleWrite("Error: Unable to read file or file is empty!" & @CRLF)
        SplashOff()
        Return False
    EndIf

    ; Split into lines
    Local $aLines = StringSplit($sFileContent, @CRLF, 1)
    If $aLines[0] < 2 Then ; Need at least header + 1 data row
        ConsoleWrite("Error: File has insufficient data" & @CRLF)
        SplashOff()
        Return False
    EndIf

    ; Process header row to find column indices
    Local $aHeaders = StringSplit($aLines[1], ",", 2)

    Local $iSessionIndex = -1
    Local $iTargetIndex = -1
    Local $iLatencyIndex = -1
    Local $iUseItIndex = -1

    For $i = 0 To UBound($aHeaders) - 1
        Local $sHeader = StringStripWS(StringReplace(StringReplace($aHeaders[$i], '"', ''), "'", ''), 3)
        Switch $sHeader
            Case "Session_Number"
                $iSessionIndex = $i
            Case "Target"
                $iTargetIndex = $i
            Case "Latency"
                $iLatencyIndex = $i
            Case "Use_It?"
                $iUseItIndex = $i
        EndSwitch
    Next

    ; Check if we found all required columns
    If $iSessionIndex = -1 Or $iTargetIndex = -1 Or $iLatencyIndex = -1 Or $iUseItIndex = -1 Then
        ConsoleWrite("Error: Could not find all required columns in the CSV!" & @CRLF)
        SplashOff()
        Return False
    EndIf

    ; Find the last session number
    Local $iMaxSessionNumber = 0
    For $i = $aLines[0] To 2 Step -1
        If $aLines[$i] <> "" Then
            Local $aFields = StringSplit($aLines[$i], ",", 2)
            If UBound($aFields) > $iSessionIndex Then
                ; Handle both single and double quotes
                Local $sSessionStr = StringReplace(StringReplace(StringStripWS($aFields[$iSessionIndex], 3), '"', ''), "'", '')
                $iMaxSessionNumber = Number($sSessionStr)
                ExitLoop
            EndIf
        EndIf
    Next

    ; Calculate the minimum session number based on max sessions
    Local $iMinSessionNumber = $iMaxSessionNumber - ($iMaxSessions - 1)

    ; Find the highest index needed to check if rows have enough columns
    Local $iMaxColIndex = $iSessionIndex
    If $iTargetIndex > $iMaxColIndex Then $iMaxColIndex = $iTargetIndex
    If $iLatencyIndex > $iMaxColIndex Then $iMaxColIndex = $iLatencyIndex
    If $iUseItIndex > $iMaxColIndex Then $iMaxColIndex = $iUseItIndex

    ; First pass: count valid entries
    Local $iValidEntryCount = 0

    For $i = 2 To $aLines[0]
        If $aLines[$i] = "" Then ContinueLoop

        Local $aFields = StringSplit($aLines[$i], ",", 2)

        ; Skip if not enough fields
        If UBound($aFields) <= $iMaxColIndex Then ContinueLoop

        ; Clean and check session number
        Local $sSessionStr = StringReplace(StringReplace(StringStripWS($aFields[$iSessionIndex], 3), '"', ''), "'", '')
        Local $iSessionNum = Number($sSessionStr)
        If $iSessionNum < $iMinSessionNumber Then ContinueLoop

        ; Check Use_It flag
        Local $sUseItStr = StringReplace(StringReplace(StringStripWS($aFields[$iUseItIndex], 3), '"', ''), "'", '')
        Local $iUseIt = Number($sUseItStr)
        If $iUseIt = 1 Then
            $iValidEntryCount += 1
        EndIf
    Next

    ; Check if we found any valid data
    If $iValidEntryCount = 0 Then
        ConsoleWrite("Error: No valid data found in the specified session range" & @CRLF)
        SplashOff()
        Return False
    EndIf

    ; Create array of exact size
    Local $aSessionData[$iValidEntryCount][3]
    Local $iIndex = 0

    ; Variables for calculating average reaction time
    Local $fTotalLatency = 0
    Local $iLatencyCount = 0

    ; Second pass: fill the array
    For $i = 2 To $aLines[0]
        If $aLines[$i] = "" Then ContinueLoop

        Local $aFields = StringSplit($aLines[$i], ",", 2)

        ; Skip if not enough fields
        If UBound($aFields) <= $iMaxColIndex Then ContinueLoop

        ; Clean and check session number
        Local $sSessionStr = StringReplace(StringReplace(StringStripWS($aFields[$iSessionIndex], 3), '"', ''), "'", '')
        Local $iSessionNum = Number($sSessionStr)
        If $iSessionNum < $iMinSessionNumber Then ContinueLoop

        ; Check Use_It flag
        Local $sUseItStr = StringReplace(StringReplace(StringStripWS($aFields[$iUseItIndex], 3), '"', ''), "'", '')
        Local $iUseIt = Number($sUseItStr)
        If $iUseIt = 1 Then
            ; Get target and latency
            Local $sTargetStr = StringReplace(StringReplace(StringStripWS($aFields[$iTargetIndex], 3), '"', ''), "'", '')
            Local $iTarget = Number($sTargetStr)

            Local $sLatencyStr = StringReplace(StringReplace(StringStripWS($aFields[$iLatencyIndex], 3), '"', ''), "'", '')
            Local $iLatency = Number($sLatencyStr)

            ; Add to total for average calculation
            $fTotalLatency += $iLatency
            $iLatencyCount += 1

            ; Store data
            $aSessionData[$iIndex][0] = $iTarget
            $aSessionData[$iIndex][1] = $iLatency
            $aSessionData[$iIndex][2] = $iSessionNum
            $iIndex += 1
        EndIf
    Next

    ; Calculate average reaction time
    If $iLatencyCount > 0 Then
        $g_iAvgReactionTimePastTwentySessions = Round($fTotalLatency / $iLatencyCount, 0)
    EndIf

    ; Store in the global variable
    $g_aSessionData = $aSessionData
    SplashOff()
    Return True
EndFunc ;_GetDataFromPastSessions


Func _ReadParametersFromINIFile(ByRef $sOutputFileName, ByRef $iMinTime, ByRef $iMaxTime, ByRef $iMinAllowedLatency, ByRef $iMaxAllowedLatency, ByRef $sOutputType, ByRef $fFeedbackDisplayTime)
    Local $hInputFile = FileOpen("react-ini.txt", 0)
    Local $sLine
    Local $asTemp

    If $hInputFile = -1 Then
        MsgBox(0, "New installation", "Detected new installation. Setting parameters to initial default values.")
        _UseDefaultParameters($sOutputFileName, $iMinTime, $iMaxTime, $iMinAllowedLatency, $iMaxAllowedLatency, $sOutputType, $fFeedbackDisplayTime)
    Else
        $sLine = FileReadLine($hInputFile)
        $asTemp = StringSplit($sLine, ";")
        $sOutputFileName = $asTemp[1]

        $sLine = FileReadLine($hInputFile)
        $asTemp = StringSplit($sLine, ";")
        $iMinTime = $asTemp[1]

        $sLine = FileReadLine($hInputFile)
        $asTemp = StringSplit($sLine, ";")
        $iMaxTime = $asTemp[1]

        $sLine = FileReadLine($hInputFile)
        $asTemp = StringSplit($sLine, ";")
        $iMinAllowedLatency = $asTemp[1]

        $sLine = FileReadLine($hInputFile)
        $asTemp = StringSplit($sLine, ";")
        $iMaxAllowedLatency = $asTemp[1]

        $sLine = FileReadLine($hInputFile)
        $asTemp = StringSplit($sLine, ";")
        $sOutputType = $asTemp[1]

        $sLine = FileReadLine($hInputFile)
		$asTemp = StringSplit($sLine, ";")
		$fFeedbackDisplayTime = $asTemp[1]
        FileClose($hInputFile)
    EndIf
EndFunc


Func _UseDefaultParameters(ByRef $sOutputFileName, ByRef $iMinTime, ByRef $iMaxTime, ByRef $iMinAllowedLatency, ByRef $iMaxAllowedLatency, ByRef $sOutputType, ByRef $fFeedbackDisplayTime)
    $sOutputFileName = @ScriptDir & "\" & "reaction-times.csv"
    $iMinTime = 450 ; Minimum wait time before target is displayed
    $iMaxTime = 700 ; Maximum wait time before target is displayed
    $iMinAllowedLatency = 200  ; Minimum allowed reaction time
    $iMaxAllowedLatency = 420  ; Maximum allowed reaction time
    $sOutputType = "WriteToFile"
    $fFeedbackDisplayTime = 1200  ; Default feedback display time in milliseconds
EndFunc


Func _UpdateINIFile($sOutputFileName, $iMinTime, $iMaxTime, $iMinAllowedLatency, $iMaxAllowedLatency, $sOutputType, $fFeedbackDisplayTime)
    Local $hOutputFile = FileOpen("react-ini.txt", 2)
    If $hOutputFile = -1 Then
        MsgBox(0, "Error", "Unable to update INI file.")
        Return (0)
    EndIf
    FileWrite($hOutputFile, $sOutputFileName & "; -- Name of output file" & @CRLF)
    FileWrite($hOutputFile, $iMinTime & "; -- Lower limit (in milliseconds) for random delay between trials" & @CRLF)
    FileWrite($hOutputFile, $iMaxTime & "; -- Upper limit (in milliseconds) for random delay between trials" & @CRLF)
    FileWrite($hOutputFile, $iMinAllowedLatency & "; -- Minimum allowed response time (ms)" & @CRLF)
    FileWrite($hOutputFile, $iMaxAllowedLatency & "; -- Maximum allowed response time (ms)" & @CRLF)
    FileWrite($hOutputFile, $sOutputType & "; -- Output type (WriteToFile or DisplayOnly)" & @CRLF)
    FileWrite($hOutputFile, $fFeedbackDisplayTime & "; -- Feedback display time (ms)" & @CRLF)
    FileClose($hOutputFile)
EndFunc


Func _ConfirmParametersWithUser(ByRef $sOutputFileName, ByRef $iMinTime, ByRef $iMaxTime, ByRef $iMinAllowedLatency, ByRef $iMaxAllowedLatency, ByRef $sOutputType, ByRef $sCondition, ByRef $sSessionCode, ByRef $fFeedbackDisplayTime)
    Local $hSetParametersForm
    Local $hRadio1
    Local $hRadio2
    Local $hOutputFileName
    Local $hChooseOutputFile
    Local $hTime1
    Local $hTime2
    Local $hLatency1
    Local $hLatency2
    Local $hFeedbackTime
    Local $hCondition
    Local $hSessionCode
    Local $hExitButton
    Local $hHelpButton
    Local $hContinueButton
    Local $nMsg
    Local $sTemp

    $hSetParametersForm = GUICreate("Reaction Time Program: Set Parameters", 580, 665, -1, -1)
    GUICtrlCreateLabel("Reaction Time Program", 175, 3, 205, 23)
    GUICtrlSetFont(-1, 14, 400, 0, "MS Sans Serif")

    ; Random delay settings
    GUICtrlCreateGroup("Random delay between trials (in milliseconds): ", 40, 40, 349, 91)
    GUICtrlSetFont(-1, 12, 400, 0, "MS Sans Serif")

    GUICtrlCreateLabel("Minimum delay:", 64, 72, 97, 20)
    GUICtrlSetFont(-1, 10, 400, 0, "MS Sans Serif")
    GUICtrlCreateLabel("Maximum delay:", 64, 104, 104, 20)
    GUICtrlSetFont(-1, 10, 400, 0, "MS Sans Serif")
    $hTime1 = GUICtrlCreateInput(450, 176, 70, 41, 21, BitOR($GUI_SS_DEFAULT_INPUT, $ES_NUMBER))
    $hTime2 = GUICtrlCreateInput(700, 176, 102, 41, 21, BitOR($GUI_SS_DEFAULT_INPUT, $ES_NUMBER))

    ; Response time limits
    GUICtrlCreateGroup("Valid response time range (in milliseconds): ", 40, 140, 349, 91)
    GUICtrlSetFont(-1, 12, 400, 0, "MS Sans Serif")

    GUICtrlCreateLabel("Minimum time:", 64, 172, 97, 20)
    GUICtrlSetFont(-1, 10, 400, 0, "MS Sans Serif")
    GUICtrlCreateLabel("Maximum time:", 64, 204, 104, 20)
    GUICtrlSetFont(-1, 10, 400, 0, "MS Sans Serif")
    $hLatency1 = GUICtrlCreateInput($iMinAllowedLatency, 176, 170, 41, 21, BitOR($GUI_SS_DEFAULT_INPUT, $ES_NUMBER))
    $hLatency2 = GUICtrlCreateInput($iMaxAllowedLatency, 176, 202, 41, 21, BitOR($GUI_SS_DEFAULT_INPUT, $ES_NUMBER))

    ; Feedback display time
    GUICtrlCreateGroup("Feedback display time (in milliseconds): ", 40, 240, 349, 70)
    GUICtrlSetFont(-1, 12, 400, 0, "MS Sans Serif")

    GUICtrlCreateLabel("Display time:", 64, 272, 97, 20)
    GUICtrlSetFont(-1, 10, 400, 0, "MS Sans Serif")
    $hFeedbackTime = GUICtrlCreateInput($fFeedbackDisplayTime, 176, 270, 70, 21, BitOR($GUI_SS_DEFAULT_INPUT, $ES_NUMBER))

    ; Output options
    GUICtrlCreateGroup("Upon completion of the testing session: ", 40, 320, 528, 110)
    GUICtrlSetFont(-1, 12, 400, 0, "MS Sans Serif")
    $hRadio1 = GUICtrlCreateRadio("Display summary and append data to output file: ", 64, 349, 422, 17)
    If $sOutputType = "WriteToFile" Then GUICtrlSetState(-1, $GUI_CHECKED)
    GUICtrlSetFont(-1, 10, 400, 0, "MS Sans Serif")
    $hOutputFileName = GUICtrlCreateInput($sOutputFileName & "  ", 81, 374, 380, 20)
    GUICtrlSetFont(-1, 10, 400, 0, "MS Sans Serif")
    GUICtrlSetState($hOutputFileName, $GUI_FOCUS)
    Send("{end}")
    $hChooseOutputFile = GUICtrlCreateButton("Choose other file...", 465, 371)
    If $sOutputType = "WriteToFile" Then
        GUICtrlSetState($hChooseOutputFile, $GUI_ENABLE)
    Else
        GUICtrlSetState($hChooseOutputFile, $GUI_DISABLE)
    EndIf
    $hRadio2 = GUICtrlCreateRadio("Display summary but do not write data to output file", 64, 403, 320, 17)
    GUICtrlSetFont(-1, 10, 400, 0, "MS Sans Serif")
    If $sOutputType = "DisplayOnly" Then GUICtrlSetState(-1, $GUI_CHECKED)

    ; Condition input
    GUICtrlCreateGroup("Please describe the condition for this testing session:", 40, 440, 520, 85)
    GUICtrlSetFont(-1, 12, 400, 0, "MS Sans Serif")
    $hCondition = GUICtrlCreateInput("", 54, 470, 500, 20)
    GUICtrlCreateLabel("Example: ""45 mins. after taking 3 tablespoons of flaxseed oil""", 56, 496, 550, 28)
    GUICtrlSetFont(-1, 10, 400, 0, "MS Sans Serif")

    ; Session Code input
    GUICtrlCreateGroup("Optional: Enter a short code for this session:", 40, 535, 520, 85)
    GUICtrlSetFont(-1, 12, 400, 0, "MS Sans Serif")
    $hSessionCode = GUICtrlCreateInput("", 54, 565, 500, 20)
    GUICtrlCreateLabel("May be useful for later data analysis (Example: ""AM-Flax-Oil"")", 56, 591, 550, 28)
    GUICtrlSetFont(-1, 10, 400, 0, "MS Sans Serif")

    ; Buttons
    $hHelpButton = GUICtrlCreateButton("Help", 272, 628, 75, 25)
    $hContinueButton = GUICtrlCreateButton("Continue", 382, 628, 75, 25, $BS_DEFPUSHBUTTON)
    $hExitButton = GUICtrlCreateButton("Exit Program", 488, 628, 75, 25)
    GUICtrlSetState($hCondition, $GUI_FOCUS)
    GUISetState(@SW_SHOW)

    While 1
        $nMsg = GUIGetMsg()
        Switch $nMsg
            Case $GUI_EVENT_CLOSE
                Exit
            Case $hRadio1
                GUICtrlSetState($hChooseOutputFile, $GUI_ENABLE)
            Case $hRadio2
                GUICtrlSetState($hChooseOutputFile, $GUI_DISABLE)
            Case $hChooseOutputFile
                $sTemp = FileOpenDialog("Choose output file", @ScriptDir & "\", "CSV and Text (*.csv;*.txt)|All files (*.*)", 2 + 8)
                If $sTemp Then
                    $sOutputFileName = $sTemp
                    GUICtrlSetData($hOutputFileName, $sOutputFileName & " ")
                EndIf
                GUICtrlSetState($hCondition, $GUI_FOCUS)
            Case $hHelpButton
                ShellExecute("http://www.astrocyte-design.com/reaction-time/index.html")
            Case $hContinueButton
                ExitLoop
            Case $hExitButton
                Exit
        EndSwitch
    WEnd

    $iMinTime = GUICtrlRead($hTime1)
    $iMaxTime = GUICtrlRead($hTime2)
    $iMinAllowedLatency = GUICtrlRead($hLatency1)
    $iMaxAllowedLatency = GUICtrlRead($hLatency2)
    $fFeedbackDisplayTime = GUICtrlRead($hFeedbackTime)
    $sOutputFileName = GUICtrlRead($hOutputFileName)
    $sOutputFileName = StringStripWS($sOutputFileName, 2)

    If (BitAND(GUICtrlRead($hRadio1), $GUI_CHECKED) = $GUI_CHECKED) Then
        $sOutputType = "WriteToFile"
    Else
        $sOutputType = "DisplayOnly"
    EndIf
    $sCondition = GUICtrlRead($hCondition)
    $sCondition = StringReplace($sCondition, ",", "") ; Remove any commas
    $sSessionCode = GUICtrlRead($hSessionCode)
    $sSessionCode = StringReplace($sSessionCode, ",", "") ; Remove any commas
    GUIDelete($hSetParametersForm)
EndFunc


Func _GetNumberOfPastSessions($sOutputFileName)
	Local $sInputFileName = $sOutputFileName
	Local $sFileContent
	Local $iSessionColIndex
	Local $aFields
	Local $iSessionNumber
	Local $iUniqueCount
	Local $aLines
	; Check if file exists
	If Not FileExists($sInputFileName) Then
		Return -1
	EndIf
	$sFileContent = FileRead($sInputFileName)
	; Check if file read was successful
	If @error Then
		Return -1
	EndIf
	; Create a more efficient storage for unique session numbers using a Map approach
	Local $oUniqueMap = ObjCreate("Scripting.Dictionary")
	; Check if dictionary object was created successfully
	If Not IsObj($oUniqueMap) Then
		Return -1
	EndIf
	$aLines = StringSplit($sFileContent, @CRLF, 1)
	; Check if file has at least a header and one data row
	If $aLines[0] < 2 Then
		Return -1
	EndIf
	; We're going to assume that Session_Number is always in the first column of the CSV file
	$iSessionColIndex = 1
	; Process each line (skip header)
	For $i = 2 To $aLines[0]
		; Skip empty lines
		If $aLines[$i] = "" Then ContinueLoop
		; Get fields from the current line
		$aFields = StringSplit($aLines[$i], ",", 1)
		; Skip if not enough fields
		If $aFields[0] < $iSessionColIndex Then ContinueLoop
		; Get Session_Number
		$iSessionNumber = $aFields[$iSessionColIndex]
		; Add to dictionary if not already there
		If Not $oUniqueMap.Exists($iSessionNumber) Then
			$oUniqueMap.Add($iSessionNumber, 1)
		EndIf
	Next
	; Get results count
	$iUniqueCount = $oUniqueMap.Count
	Return($iUniqueCount)
EndFunc


Func _GetLatencyPercentile($iCurrentLatency, $iCurrentTarget)
	; Fast function to calculate percentile - uses the cached data
	; @param $iCurrentLatency The user's latency for the current trial
	; @param $iCurrentTarget The user's target for the current trial
	; @return Percentile rank (0-100) or -100 on error
    ; Check if data has been loaded
    If UBound($g_aSessionData) = 0 Then
        ConsoleWrite("Error: No session data has been loaded. Call LoadPastSessionsData first." & @CRLF)
        Return -100
    EndIf

    ; Calculate percentile
    Local $iCount = 0
    Local $iTotalMatchingTarget = 0

    For $i = 0 To UBound($g_aSessionData) - 1
        If $g_aSessionData[$i][0] = $iCurrentTarget Then
            $iTotalMatchingTarget += 1
            If $g_aSessionData[$i][1] <= $iCurrentLatency Then
                $iCount += 1
            EndIf
        EndIf
    Next

    ; If no matching target values found, return error
    If $iTotalMatchingTarget = 0 Then
        ConsoleWrite("Warning: No data found for target value " & $iCurrentTarget & @CRLF)
        Return -100
    EndIf

    ; Calculate percentile rank
    Return Round((1 - ($iCount / $iTotalMatchingTarget)) * 100, 0)
EndFunc


Func _PresentTargetAndGetResponse($sTarget, $hGui)
    Local $iReactionTime
    Local $iCorrect
    Local $sLocalTime
    Local $avResults[4]
	Local $hLabelMessage

    $hLabelMessage = GUICtrlCreateLabel("|", 0, 0, $iGUIWidth, $iGUIHeight - 10, BitOR($SS_CENTER, $SS_CENTERIMAGE))
    GUICtrlSetFont(-1, $iLargerFont)

    ; Make sure GUI is visible and active
    GUISetState(@SW_SHOW, $hGui)

    If ($iMinTime < $iMaxTime) Then
        Sleep(Random($iMinTime, $iMaxTime, 1))
    Else
        Sleep($iMinTime)
    EndIf

    GUICtrlDelete($hLabelMessage)

    If (StringIsDigit($sTarget) = 1) And ($sTarget >= 2) And ($sTarget <= 8) Then
        _ShowSingleDigitBetweenTwoAndEight($sTarget, $hGui, $hLabelMessage)
    Else
        $hLabelMessage = GUICtrlCreateLabel($sTarget, 0, 0, $iGUIWidth, $iGUIHeight, BitOR($SS_CENTER, $SS_CENTERIMAGE))
    EndIf

    ; Make sure GUI is still visible and active
    GUISetState(@SW_SHOW, $hGui)

    $hBegin = TimerInit()
    $iCorrect = _GetKeyPress($sTarget)
    $iReactionTime = Round(TimerDiff($hBegin))
    $sLocalTime = _GetTime()
    GUICtrlDelete($hLabelMessage)

    $avResults[0] = $sTarget
    $avResults[1] = $iReactionTime
    $avResults[2] = $iCorrect
    $avResults[3] = $sLocalTime
    Return ($avResults)
EndFunc


Func _WaitForUserToHitAFive($hGui, $sTextToDisplay = "", $iFontSize = 20)
	Local $sText
	Local $hLabelMessage
	Local $aStringSize
	Local $iXPosition
	Local $iYPosition

	If $sTextToDisplay <> "" Then
		$sText = "Hit ''5'' to start" & $sTextToDisplay
	Else
		$sText = "Hit ''5'' to start"
	EndIf
	$aStringSize = _StringSize($sText, $iFontSize)
	; Determine where the string needs to be positioned in the GUI
	$iXPosition = ($iGUIWidth - $aStringSize[2]) / 2
	$iYPosition = ($iGUIHeight - $aStringSize[3]) / 2
	$hLabelMessage = GUICtrlCreateLabel($sText, $iXPosition, $iYPosition, $aStringSize[2], $aStringSize[3], $SS_CENTER)
	GUICtrlSetFont($hLabelMessage, $iFontSize)
	While _IsPressed(35) <> 1
	WEnd
	Sleep(50)
	GUICtrlDelete($hLabelMessage)
EndFunc   ;==>_WaitForUserToHitAFive


Func _ShowSuccessFeedback($iLatency, $iTarget)
    Local $sTempGUIMessage
    If ($bUsePercentiles = True) Then
        $iPercentileRank = _GetLatencyPercentile($iLatency, $iTarget)
        $sTempGUIMessage = " Correct!" & @CRLF & $iLatency & "ms" & " / " & $iPercentileRank & "%"
    Else
        $sTempGUIMessage = " Correct!" & @CRLF & $iLatency & "ms"
    EndIf

    Local $aStringSize = _StringSize($sTempGUIMessage, $iMediumFont)
    Local $iXPosition = ($iGUIWidth - $aStringSize[2]) / 2
    Local $iYPosition = ($iGUIHeight - $aStringSize[3]) / 2
    Local $hLabelMessage = GUICtrlCreateLabel($sTempGUIMessage, $iXPosition, $iYPosition, $aStringSize[2], $aStringSize[3], $SS_CENTER)
    GUICtrlSetFont($hLabelMessage, $iMediumFont)
    Sleep($fFeedbackDisplayTime)
    GUICtrlDelete($hLabelMessage)
EndFunc


Func _ShowFailureFeedback($avResult)
    Local $sTempGUIMessage
    Local $iLatency = $avResult[1]

    If ($avResult[2] == 0) And ($iLatency <= $iMaxAllowedLatency) And ($iLatency >= $iMinAllowedLatency) Then
        $sTempGUIMessage = "Wrong"
    ElseIf ($avResult[2] == 1) And ($iLatency > $iMaxAllowedLatency) Then
        $sTempGUIMessage = "Too slow" & @CRLF & $iLatency & "ms"
    ElseIf ($avResult[2] == 0) And ($iLatency > $iMaxAllowedLatency) Then
        $sTempGUIMessage = "Wrong and too slow" & @CRLF & $iLatency & "ms"
    ElseIf ($avResult[2] == 1) And ($iLatency < $iMinAllowedLatency) Then
        $sTempGUIMessage = "Too fast" & @CRLF & $iLatency & "ms"
    ElseIf ($avResult[2] == 0) And ($iLatency < $iMinAllowedLatency) Then
        $sTempGUIMessage = "Wrong and too fast" & @CRLF & $iLatency & "ms"
    EndIf

    Local $aStringSize = _StringSize($sTempGUIMessage, $iMediumFont)
    Local $iXPosition = ($iGUIWidth - $aStringSize[2]) / 2
    Local $iYPosition = ($iGUIHeight - $aStringSize[3]) / 2
    Local $hLabelMessage = GUICtrlCreateLabel($sTempGUIMessage, $iXPosition, $iYPosition, $aStringSize[2], $aStringSize[3], $SS_CENTER)
    GUICtrlSetFont($hLabelMessage, $iMediumFont)
    Sleep($iDisplayDurationForNonRegularFeedbackGUI)
    GUICtrlDelete($hLabelMessage)
EndFunc


Func _ShowCorrectionSuccessFeedback()
    Local $sTempGUIMessage = "Correct!"
    Local $aStringSize = _StringSize($sTempGUIMessage, $iMediumFont)
    Local $iXPosition = ($iGUIWidth - $aStringSize[2]) / 2
    Local $iYPosition = ($iGUIHeight - $aStringSize[3]) / 2
    Local $hLabelMessage = GUICtrlCreateLabel($sTempGUIMessage, $iXPosition, $iYPosition, $aStringSize[2], $aStringSize[3], $SS_CENTER)
    GUICtrlSetFont($hLabelMessage, $iMediumFont)
    Sleep($iDisplayDurationForNonRegularFeedbackGUI)
    GUICtrlDelete($hLabelMessage)
EndFunc


Func _ShowCorrectionFailureFeedback()
    Local $sTempGUIMessage = "Wrong. Try again."
    Local $aStringSize = _StringSize($sTempGUIMessage, $iMediumFont)
    Local $iXPosition = ($iGUIWidth - $aStringSize[2]) / 2
    Local $iYPosition = ($iGUIHeight - $aStringSize[3]) / 2
    Local $hLabelMessage = GUICtrlCreateLabel($sTempGUIMessage, $iXPosition, $iYPosition, $aStringSize[2], $aStringSize[3], $SS_CENTER)
    GUICtrlSetFont($hLabelMessage, $iMediumFont)
    Sleep($iDisplayDurationForNonRegularFeedbackGUI)
    GUICtrlDelete($hLabelMessage)
EndFunc


Func _ShowSingleDigitBetweenTwoAndEight($sTarget, $hGui, ByRef $hLabelMessage)
	Switch Int($sTarget)
		Case 2
			$hLabelMessage = GUICtrlCreateLabel("2", 10, 0, $iGUIWidth, $iGUIHeight, $SS_CENTERIMAGE)
			GUICtrlSetFont(-1, $iLargerFont)
		Case 3
			$hLabelMessage = GUICtrlCreateLabel("3", (10 + ($iGUIWidth / 7)), 0, $iGUIWidth, $iGUIHeight, $SS_CENTERIMAGE)
			GUICtrlSetFont(-1, $iLargerFont)
		Case 4
			$hLabelMessage = GUICtrlCreateLabel("4", (10 + 2 * ($iGUIWidth / 7)), 0, $iGUIWidth, $iGUIHeight, $SS_CENTERIMAGE)
			GUICtrlSetFont(-1, $iLargerFont)
		Case 5
			$hLabelMessage = GUICtrlCreateLabel("5", (10 + 3 * ($iGUIWidth / 7)), 0, $iGUIWidth, $iGUIHeight, $SS_CENTERIMAGE)
			GUICtrlSetFont(-1, $iLargerFont)
		Case 6
			$hLabelMessage = GUICtrlCreateLabel("6", (10 + 4 * ($iGUIWidth / 7)), 0, $iGUIWidth, $iGUIHeight, $SS_CENTERIMAGE)
			GUICtrlSetFont(-1, $iLargerFont)
		Case 7
			$hLabelMessage = GUICtrlCreateLabel("7", (10 + 5 * ($iGUIWidth / 7)), 0, $iGUIWidth, $iGUIHeight, $SS_CENTERIMAGE)
			GUICtrlSetFont(-1, $iLargerFont)
		Case 8
			$hLabelMessage = GUICtrlCreateLabel("8", (10 + 6 * ($iGUIWidth / 7)), 0, $iGUIWidth, $iGUIHeight, $SS_CENTERIMAGE)
			GUICtrlSetFont(-1, $iLargerFont)
	EndSwitch
EndFunc   ;==>_ShowSingleDigitBetweenTwoAndEight


Func _GetKeyPress($iRightAnswer)
	Local $iRightAnswerCode = 30 + $iRightAnswer ; Convert the answer into the ASCII code that's required by the _IsPressed function
	Local $iWrongAnswers[10] = [30, 31, 32, 33, 34, 35, 36, 37, 38, 39] ; Initialize array of wrong answer digits
	$iWrongAnswers[$iRightAnswer] = '20' ; Replace the right answer with the code for the space character
	While 1
		For $x = 0 To 9
			If _IsPressed($iRightAnswerCode) Then Return (1)
			If _IsPressed($iWrongAnswers[$x]) Then Return (0)
		Next
	WEnd
EndFunc   ;==>_GetKeyPress


Func _DisplaySummary($iNumberOfTargets, $iTotalTrialsConducted, $avReactionTimeData)
	Local $iMaxReactionTime = -99999
	Local $iMinReactionTime = 99999
	Local $iNumAnsweredCorrectly = 0
	Local $iSumOfTimes = 0
	Local $iNumWrong = 0
	Local $iAvgTime
	Local $sTemp
	For $x = 0 To ($iTotalTrialsConducted - 1)
		If ($avReactionTimeData[$x][2] = 1) And ($avReactionTimeData[$x][1] > $iMaxReactionTime) And ($avReactionTimeData[$x][0] <> "c") And ($avReactionTimeData[$x][1] <= $iMaxAllowedLatency) And ($avReactionTimeData[$x][1] >= $iMinAllowedLatency) Then
			$iMaxReactionTime = $avReactionTimeData[$x][1]
		EndIf
		If ($avReactionTimeData[$x][2] = 1) And ($avReactionTimeData[$x][1] < $iMinReactionTime) And ($avReactionTimeData[$x][0] <> "c") And ($avReactionTimeData[$x][1] <= $iMaxAllowedLatency) And ($avReactionTimeData[$x][1] >= $iMinAllowedLatency) Then
			$iMinReactionTime = $avReactionTimeData[$x][1]
		EndIf
		If ($avReactionTimeData[$x][2] = 0) And ($avReactionTimeData[$x][0] <> "c") And ($avReactionTimeData[$x][1] < $iMaxAllowedLatency) And ($avReactionTimeData[$x][1] > $iMinAllowedLatency) Then
			$iNumWrong = $iNumWrong + 1 ; Don't include mistakes made during correction trials, and don't include answers that were right but were either too slow or too fast
		EndIf
		If $avReactionTimeData[$x][2] = 1 And $avReactionTimeData[$x][0] <> "c" And ($avReactionTimeData[$x][1] <= $iMaxAllowedLatency) And ($avReactionTimeData[$x][1] >= $iMinAllowedLatency) Then
			$iNumAnsweredCorrectly = $iNumAnsweredCorrectly + 1 ; This is really just a safety check, since the expectation is that the number of correct (valid) answers should be equal to $iNumberOfTargets
			$iSumOfTimes = $iSumOfTimes + $avReactionTimeData[$x][1] ; Don't include reaction times from correction trials or trials that were either too slow or too fast
		EndIf
	Next
	$iAvgTime = Round($iSumOfTimes / $iNumberOfTargets)
	$sTemp = "Average time = " & $iAvgTime & " ms" & @CRLF
	If $g_iAvgReactionTimePastTwentySessions > 0 Then
		$sTemp = $sTemp & "Avg. of last 20 sessions = " & $g_iAvgReactionTimePastTwentySessions & " ms" & @CRLF
	EndIf
	$sTemp = $sTemp & "Fastest time = " & $iMinReactionTime & " ms" & @CRLF & "Slowest time = "
	$sTemp = $sTemp & $iMaxReactionTime & " ms" & @CRLF & "Number correct = " & $iNumAnsweredCorrectly & @CRLF & "Number incorrect = "
	$sTemp = $sTemp & $iNumWrong & @CRLF & "Percent incorrect = " & Int(100 * ($iNumWrong / $iNumberOfTargets)) & "%"
	MsgBox(0, "Summary", $sTemp)
EndFunc   ;==>_DisplaySummary


Func _CheckForValidResults($iNumberOfTargets, $iTotalTrialsConducted, $avReactionTimeData)
	; Before we actually write the data to the data file, we want to be sure that our results conform to our expectations.
	; (in the sense that there are the expected number of valid trials).  In the typical case I've been using, there
	; should be 35 valid trials at the end of a testing session.
	Local $iCounter = 0
	Local $sMsgBoxText
	; Count the number of valid answers
	For $x = 0 To ($iTotalTrialsConducted - 1)
		If ($avReactionTimeData[$x][0] <> "c") And ($avReactionTimeData[$x][2] = 1) And ($avReactionTimeData[$x][1] <= $iMaxAllowedLatency) And ($avReactionTimeData[$x][1] >= $iMinAllowedLatency) Then
			$iCounter = $iCounter + 1
		EndIf
	Next
	If $iCounter <> $iNumberOfTargets Then
		$sMsgBoxText = "ERROR: There were " & $iCounter & " valid trials, when we expected " & $iNumberOfTargets &". Something went wrong! Debugging is needed."
		Return False
	Else
		Return True
	EndIf
EndFunc   ;==>_CheckForValidResults


Func _WriteReactionTimeData($sCondition, $iTotalTrialsConducted, $avReactionTimeData)
	Local $hOutputFile
	Local $iSessionNumber
	$iSessionNumber = _GetLastSessionNumber($sOutputFileName) + 1
	$hOutputFile = FileOpen($sOutputFileName, 1)
	; Check if file opened for writing OK
	If $hOutputFile = -1 Then
		MsgBox(0, "Error", "Error: Unable to open output file.")
		Return (0)
	EndIf
	; MsgBox(0, "", "$iSessionNumber = " & $iSessionNumber)
	; Check to see if this is a new output file. If so, output the column headings:
	If $iSessionNumber = 1 Then
		FileWrite($hOutputFile, "Session_Number,Condition,Session_Code,Latency,Target,Correct?,Use_It?,Local_Time" & @CRLF)
	EndIf
	For $x = 0 To ($iTotalTrialsConducted - 1)
		FileWrite($hOutputFile, $iSessionNumber & ",")      ; Write the session number
		FileWrite($hOutputFile, $sCondition & ",")          ; Write the condition
		FileWrite($hOutputFile, $sSessionCode & ",")        ; Write the session code
		FileWrite($hOutputFile, $avReactionTimeData[$x][1] & ",") ; Write the user's reaction time for that particular target
		FileWrite($hOutputFile, "'" & $avReactionTimeData[$x][0] & "'" & ",") ; Write the specific target that was presented to the user
		FileWrite($hOutputFile, $avReactionTimeData[$x][2] & ",") ; Write a '1' if the user pressed the correct key, or a '0' if the user hit the wrong key
		If ($avReactionTimeData[$x][2] = 1 And $avReactionTimeData[$x][0] <> 'c' And $avReactionTimeData[$x][1] <= $iMaxAllowedLatency And $avReactionTimeData[$x][1] >= $iMinAllowedLatency) Then ; It was correct, not a correction trial, and was responded to within a reasonable amount of time (latency)
			FileWrite($hOutputFile, 1 & ",") ; Write a '1' if this trial should be used for subsequent data analysis
		Else
			FileWrite($hOutputFile, 0 & ",") ; Write a '0' if this trial should NOT be used for subsequent data analysis
		EndIf
		FileWrite($hOutputFile, $avReactionTimeData[$x][3] & @CRLF) ; Write the local time at which the user responded to the target (followed by a carriage return / line feed)
	Next
	FileClose($hOutputFile)
EndFunc   ;==>_WriteReactionTimeData


Func _GetLastSessionNumber($sFileName)
	Local $hInputFile = FileOpen($sFileName, 0)
	Local $sLastLine
	Local $asTemp
	; If the file doesn't exist, return a zero
	If $hInputFile = -1 Then
		FileClose($hInputFile)
		Return (0)
	Else
		$sLastLine = FileReadLine($hInputFile, -1) ; Read the last line of the output file
		If (@error = -1) Or (@error = 1) Then Return (0)
		$asTemp = StringSplit($sLastLine, ",") ; Get the first piece of data from the last line (i.e., the number of sessions thus far)
		FileClose($hInputFile)
		; Check to see if there is a valid prior session number
		If ($asTemp[1] = "") Or (Not StringIsInt($asTemp[1])) Then
			Return (0)
		Else
			Return ($asTemp[1])
		EndIf
	EndIf
EndFunc   ;==>_GetLastSessionNumber


Func _Terminate()
	Exit 0
EndFunc   ;==>_Terminate


Func _GetTime()
	Local $tLocal = _Date_Time_GetLocalTime()
	Return (_Date_Time_SystemTimeToDateTimeStr($tLocal))
EndFunc   ;==>_GetTime
